var EventEmitter = require('events').EventEmitter
var inherits = require('util').inherits

inherits(Channel, EventEmitter)

module.exports = Channel

function Channel(options) {
  if (!(this instanceof Channel))
    return new Channel(options)

  options = options || {}

  EventEmitter.call(this, options)

  this.concurrency = options.concurrency || Infinity
  // closed by default
  this.closed = !(options.closed === false || options.open)
  this.discard = options.discard

  this.fns = []
  this.results = []

  this.fnIndex = 0
  this.resultIndex = 0
  this.pending = 0
  this.errors = 0

  this.reading = true

  this.next = this.next.bind(this)
}

Object.defineProperty(Channel.prototype, 'pushable', {
  get: function () {
    return this.reading && this.pending < this.concurrency
  }
})

// read queue
Object.defineProperty(Channel.prototype, 'queue', {
  get: function () {
    var queue = this.results.length + this.fns.length
    if (this.discard) queue += this.pending
    return queue
  }
})

// you can read from a channel if there's a read queue or if this channel is not closed
Object.defineProperty(Channel.prototype, 'readable', {
  get: function () {
    return this.queue || !this.closed
  }
})

Channel.prototype.open = function () {
  this.closed = false
  this.emit('open')
  return this
}

Channel.prototype.close = function () {
  this.closed = true
  this.emit('close')
  return this
}

// when the channel is open,
// wait for the first push event (returns true)
// or close event (returns false)
Channel.prototype.pushed = wrapGenerator.mark(function() {
  var self;

  return wrapGenerator(function($ctx0) {
    while (1) switch ($ctx0.next) {
    case 0:
      if (!this.closed) {
        $ctx0.next = 5;
        break;
      }

      $ctx0.rval = false;
      delete $ctx0.thrown;
      $ctx0.next = 17;
      break;
    case 5:
      if (!this.queue) {
        $ctx0.next = 10;
        break;
      }

      $ctx0.rval = true;
      delete $ctx0.thrown;
      $ctx0.next = 17;
      break;
    case 10:
      self = this;
      $ctx0.next = 13;

      return function (done) {
        self.on('close', close)
        self.on('push', push)

        function close() {
          cleanup()
          done(null, false)
        }

        function push() {
          cleanup()
          done(null, true)
        }

        function cleanup() {
          self.removeListener('close', close)
          self.removeListener('push', push)
        }
      }
    case 13:
      $ctx0.rval = $ctx0.sent;
      delete $ctx0.thrown;
      $ctx0.next = 17;
      break;
    case 17:
    case "end":
      return $ctx0.stop();
    }
  }, this);
})

/**
 * Push a function to the channel.
 * If `null`, just means closing the channel.
 */

Channel.prototype.push = function (fn) {
  if (fn == null) return this.close()

  if (typeof fn !== 'function')
    throw new TypeError('you may only push functions')

  this.fns.push(fn)
  if (!this.discard) this.results.length++
  this.call()
  this.emit('push')
  return this
}

Channel.prototype.call = function () {
  if (!this.pushable) return
  if (!this.fns.length) return

  var fn = this.fns.shift()
  var index = this.fnIndex++
  this.pending++

  var self = this
  fn(function (err, res) {
    self.pending--
    if (err) {
      self.reading = false
      self.errors++
      if (self.discard) {
        self.results.push(err)
      } else {
        self.results[index - self.resultIndex] = err
      }
    } else if (!self.discard) {
      self.results[index - self.resultIndex] = arguments.length > 2
        ? [].slice.call(arguments, 1)
        : res
    }

    self.emit(String(index))
    self.emit('callback')
    self.call()
  })
}

Channel.prototype.read = wrapGenerator.mark(function() {
  var self, res;

  return wrapGenerator(function($ctx1) {
    while (1) switch ($ctx1.next) {
    case 0:
      self = this;

      // continue executing callbacks if no errors occured
      if (!this.reading && !this.errors) this.reading = true;

      if (!!self.discard) {
        $ctx1.next = 15;
        break;
      }

      if (!(self.results.length && 0 in self.results)) {
        $ctx1.next = 13;
        break;
      }

      res = self.results.shift();
      self.resultIndex++;

      if (!(res instanceof Error)) {
        $ctx1.next = 9;
        break;
      }

      self.errors--;
      throw res;
    case 9:
      $ctx1.rval = res;
      delete $ctx1.thrown;
      $ctx1.next = 24;
      break;
    case 13:
      $ctx1.next = 18;
      break;
    case 15:
      if (!self.results.length) {
        $ctx1.next = 18;
        break;
      }

      // these can only be errors
      self.errors--;

      throw self.results.shift();
    case 18:
      $ctx1.next = 20;
      return this.next;
    case 20:
      $ctx1.rval = $ctx1.sent;
      delete $ctx1.thrown;
      $ctx1.next = 24;
      break;
    case 24:
    case "end":
      return $ctx1.stop();
    }
  }, this);
})

Channel.prototype.next = function (done) {
  var event = this.discard
    ? 'callback'
    : String(this.resultIndex)

  var self = this
  this.once(event, onevent)
  if (!this.closed && !this.queue) this.once('close', onclose)

  function onevent() {
    cleanup()
    if (!this.discard) {
      var res = this.results.shift()
      this.resultIndex++
      if (res instanceof Error) {
        this.errors--
        done(res)
      } else {
        done(null, res)
      }
    } else if (this.results.length) {
      // these can only be errors
      this.errors--
      done(this.results.shift())
    } else {
      done()
    }
  }

  function onclose() {
    cleanup()
    done()
  }

  function cleanup() {
    self.removeListener(event, onevent)
    self.removeListener('close', onclose)
  }
}

Channel.prototype.flush = wrapGenerator.mark(function() {
  var results;

  return wrapGenerator(function($ctx2) {
    while (1) switch ($ctx2.next) {
    case 0:
      if (!this.discard) {
        $ctx2.next = 8;
        break;
      }
    case 1:
      if (!this.readable) {
        $ctx2.next = 5;
        break;
      }

      return $ctx2.delegateYield(this.read(), "t0", 3);
    case 3:
      $ctx2.next = 1;
      break;
    case 5:
      delete $ctx2.thrown;
      $ctx2.next = 18;
      break;
    case 8:
      results = [];
    case 9:
      if (!this.readable) {
        $ctx2.next = 14;
        break;
      }

      return $ctx2.delegateYield(this.read(), "t1", 11);
    case 11:
      results.push($ctx2.t1);
      $ctx2.next = 9;
      break;
    case 14:
      $ctx2.rval = results;
      delete $ctx2.thrown;
      $ctx2.next = 18;
      break;
    case 18:
    case "end":
      return $ctx2.stop();
    }
  }, this);
})